from six import StringIO

from .representer import Representer
from .serializer import Serializer


class Dumper(Serializer, Representer):
    def __init__(self, stream):
        Serializer.__init__(self, stream)
        Representer.__init__(self)


def dump(data, stream=None, Dumper=Dumper):
    """
    Serialize Python objects into a WRF namelist stream
    :param data:
    :param stream:
    :return:
    """
    getvalue = None
    if stream is None:
        stream = StringIO()
        getvalue = stream.getvalue

    dumper = Dumper(stream)

    dumper.open()
    for d in data:
        dumper.represent(d, data[d])
    dumper.close()

    if getvalue:
        return getvalue()



class Node(object):
    def __init__(self, value, start_mark, end_mark):
        self.value = value
        self.start_mark = start_mark
        self.end_mark = end_mark

    def __repr__(self):
        return self.value


class ScalarNode(Node):
    id = 'scalar'

    def __init__(self, value, start_mark=None, end_mark=None):
        super(ScalarNode, self).__init__(value, start_mark, end_mark)


class CollectionNode(Node):
    id = 'collection'

    def __init__(self, value, start_mark=None, end_mark=None):
        super(CollectionNode, self).__init__(value, start_mark, end_mark)


class SectionNode(CollectionNode):
    id = 'section'


class SequenceNode(CollectionNode):
    id = 'sequence'



import re
from collections import defaultdict

import yaml

default_parser = re.compile('^[a-z0-9 _()\t]*?=(.+?)[;\(\|]?(.*)$', re.DOTALL)


def is_command(l):
    return is_section_item(l) or default_parser.search(l) is not None


def is_section_item(l):
    try:
        return l.strip()[0] == '&'
    except IndexError:
        return False


class ConfigItem(object):
    def __init__(self, line):
        self.raw = line
        self.name = None
        self.is_multi_dim = False
        self.is_section = False
        self.description = ""
        self.default = None
        self.process(line)

    def __str__(self):
        if self.is_section:
            return '<section {}>'.format(self.name)
        else:
            return '<item {} default={} is_multi_dim={}>'.format(self.name, self.default, self.is_multi_dim)

    def process(self, l):
        if l[0] == '&':
            self.is_section = True
            self.name = l.strip()
            if ' ' in self.name:
                self.name = self.name[:self.name.find(' ')]
            self.name = self.name.strip('&:')
        else:
            raw_name = l[:l.find('=')]
            if '(max_dom)' in raw_name:
                self.is_multi_dim = True
            self.name = raw_name.replace('(max_dom)', '').strip()

            res = default_parser.search(l, )
            if res:
                self.default = res.group(1).strip()
                self.description = res.group(2).strip()
                self.description = '\n'.join([a.strip() for a in self.description.split('\n')])
                pass

    def to_dict(self):
        return {
            'name': self.name,
            'default': self.default,
            'description': self.description,
            'is_multi_dim': self.is_multi_dim
        }


def get_next_item(lines):
    """
    Find the next item
    :param lines: List of lines
    :return: The line to process with any newlines removed
    """
    while len(lines):
        l = lines.pop(0)

        # Determine if the line is part of a var or text
        if is_command(l):
            res = l

            # Peek at the next item and see if it is not blank
            while len(lines):
                if lines[0] != "\n" and not is_command(lines[0]):
                    l = lines.pop(0)
                    res += l
                else:
                    break

            return res.strip()


def stringify(items):
    d = {}
    for item in items:
        d[item] = []
        for v in items[item]:
            d[item].append(v.to_dict())
    return yaml.dump(d, default_flow_style=False)


def process_namelist(fname):
    with open(fname) as fh:
        lines = fh.readlines()

    # Store the current section to parse the hierachy
    current_section = None
    items = defaultdict(set)
    while len(lines):
        item_str = get_next_item(lines)
        item = ConfigItem(item_str)

        if item.is_section:
            current_section = item
        else:
            if item.name and item.name not in items[current_section.name]:
                items[current_section.name].add(item)
    # Stringify
    return stringify(items)



import warnings
from collections import OrderedDict
from .nodes import ScalarNode, SectionNode, SequenceNode


class RepresenterError(Exception):
    pass


class BaseRepresenter(object):
    wrf_representers = {}

    def represent(self, section_name, data):
        node = self.represent_data(data)
        self.serialize(section_name, node)

    def represent_data(self, data, **kwargs):
        data_types = type(data).__mro__
        if data_types[0] in self.wrf_representers:
            node = self.wrf_representers[data_types[0]](self, data, **kwargs)

            return node

    @classmethod
    def add_representer(cls, data_type, representer):
        if not 'wrf_representers' in cls.__dict__:
            cls.wrf_representers = cls.wrf_representers.copy()
        cls.wrf_representers[data_type] = representer

    def represent_scalar(self, value):
        node = ScalarNode(value, end_mark=',')
        return node

    def represent_sequence(self, sequence, **kwargs):
        value = []
        node = SequenceNode(value)
        for item in sequence:
            node_item = self.represent_data(item, **kwargs)
            value.append(node_item)
        return node

    def represent_section(self, mapping):
        value = []
        node = SectionNode(value)
        if hasattr(mapping, 'items'):
            mapping = list(mapping.items())
            try:
                mapping = sorted(mapping)
            except TypeError:
                pass
        for item_key, item_value in mapping:
            node_key = self.represent_data(item_key)
            node_key.end_mark = None
            node_value = self.represent_data(item_value, is_value=True)
            value.append((node_key, node_value))
        return node


class Representer(BaseRepresenter):

    def represent_str(self, data, is_value=False):
        if ',' in data:
            warnings.warn('Value was potentially incorrectly intepreted as a string, please check: {}'.format(data))
        if is_value:
            return self.represent_scalar("'{}'".format(data))

        return self.represent_scalar(data)

    def represent_bool(self, data, **kwargs):
        if data:
            value = '.true.'
        else:
            value = '.false.'
        return self.represent_scalar(value)

    def represent_int(self, data, **kwargs):
        return self.represent_scalar(str(data))

    inf_value = 1e300
    while repr(inf_value) != repr(inf_value * inf_value):
        inf_value *= inf_value

    def represent_float(self, data, **kwargs):
        if data != data or (data == 0.0 and data == 1.0):
            value = '.nan'
        elif data == self.inf_value:
            value = '.inf'
        elif data == -self.inf_value:
            value = '-.inf'
        else:
            value = repr(data).lower()
            # Note that in some cases `repr(data)` represents a float number
            # without the decimal parts.  For instance:
            #   >>> repr(1e17)
            #   '1e17'
            # Unfortunately, this is not a valid float representation according
            # to the definition of the `!!float` tag.  We fix this by adding
            # '.0' before the 'e' symbol.
            if '.' not in value and 'e' in value:
                value = value.replace('e', '.0e', 1)
        return self.represent_scalar(value)

    def represent_list(self, data, **kwargs):
        # pairs = (len(data) > 0 and isinstance(data, list))
        # if pairs:
        #    for item in data:
        #        if not isinstance(item, tuple) or len(item) != 2:
        #            pairs = False
        #            break
        # if not pairs:
        return self.represent_sequence(data, **kwargs)

    def represent_dict(self, data, **kwargs):
        return self.represent_section(data)


Representer.add_representer(str,
                            Representer.represent_str)

Representer.add_representer(bool,
                            Representer.represent_bool)

Representer.add_representer(int,
                            Representer.represent_int)

Representer.add_representer(float,
                            Representer.represent_float)

Representer.add_representer(list,
                            Representer.represent_list)

Representer.add_representer(tuple,
                            Representer.represent_list)

Representer.add_representer(dict,
                            Representer.represent_dict)

Representer.add_representer(OrderedDict,
                            Representer.represent_dict)


from datetime import datetime


def merge_dicts(a, b):
    """
    Merge dicts a and b, with values in b overriding a

    Merges at the the second level. i.e. Merges dict top level values
    :return: New dict
    """
    c = a.copy()
    for k in b:
        if k not in c:
            c[k] = b[k]
        else:
            c[k].update(b[k])
    return c


def convert_str_to_dt(s):
    return datetime.strptime(s, '%Y-%m-%d_%H:%M:%S')

def convert_dt_to_str(dt):
    return dt.strftime('%Y-%m-%d_%H:%M:%S')


def make_list(v, num):
    return [v] * num


__all__ = ['Serializer', 'SerializerError']

from .nodes import *


class SerializerError(Exception):
    pass


class Serializer:
    def __init__(self, stream):
        self.closed = None
        self.stream = stream

    def open(self):
        if self.closed is None:
            self.closed = False
        elif self.closed:
            raise SerializerError("serializer is closed")
        else:
            raise SerializerError("serializer is already opened")

    def close(self):
        if self.closed is None:
            raise SerializerError("serializer is not opened")
        elif not self.closed:
            self.closed = True

    def serialize(self, section_name, node):
        if self.closed is None:
            raise SerializerError("serializer is not opened")
        elif self.closed:
            raise SerializerError("serializer is closed")

        self.emit('&{}\n'.format(section_name))
        self.serialize_node(node)
        self.emit('/\n\n')

    def emit(self, chars):
        if chars is not None:
            self.stream.write(chars)

    def serialize_node(self, node, indent=0):
        if isinstance(node, ScalarNode):
            if indent:
                self.emit(' ' * indent)
            self.emit(node.start_mark)
            self.emit(node.value)
            self.emit(node.end_mark)
        elif isinstance(node, SequenceNode):
            for item in node.value:
                item.start_mark = '\t'
                self.serialize_node(item)
        elif isinstance(node, SectionNode):
            for key, value in node.value:
                self.serialize_node(key, indent=2)
                self.emit(' = ')
                self.serialize_node(value)
                self.emit('\n')



from os import path

from setuptools import setup, find_packages

from wrfconf import __version__

with open(path.join(path.abspath(path.dirname(__file__)), 'README.rst')) as f:
    long_description = f.read()

setup(name='wrfconf',
      version=__version__,
      description='Configuration generator for WRF namelists',
      long_description=long_description,
      author='Jared Lewis',
      author_email='jared@bodekerscientific.com',
      license='MIT',
      keywords='wrf config generate generator science forecast',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Science/Research',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering :: Atmospheric Science'
      ],
      install_requires=[
          'pyyaml',
          'six'
      ],
      packages=find_packages(exclude='tests'),
      entry_points={
          'console_scripts':
              ['wrfconf = wrfconf.cli:main']
      },
      zip_safe=False)



import argparse

from wrfconf.conf.parse_namelist import process_namelist
from wrfconf.process import process_conf_file

def gen_params_parser(subparsers):
    gen = subparsers.add_parser('gen_params', help='Generate a configuration file containing the valid WRF parameters')
    gen.add_argument('input', help='input README.namelist filename')


def create_parser(subparsers):
    create = subparsers.add_parser('create', help='Create new configuration files for a WRF run')
    create.add_argument('input', help='YML configuration file for the run')
    create.add_argument('-n', '--namelist', default='.', help='Folder to store the WRF namelist file')
    create.add_argument('-w', '--wps', default='.', help='Folder to store the WPS file')


def run_command(args):
    if args.cmd == 'gen_params':
        print(process_namelist(args.input))
    elif args.cmd == 'create':
        process_conf_file(args.input, args.namelist, args.wps)


def main():
    parser = argparse.ArgumentParser(prog='wrfconf',
                                     description="Generate WRF configuration from structured YAML files")
    subparsers = parser.add_subparsers(dest='cmd')
    subparsers.required = True

    create_parser(subparsers)
    gen_params_parser(subparsers)

    args = parser.parse_args()
    run_command(args)


main()


from unittest import TestCase

from wrfconf.conf.parse_namelist import ConfigItem, get_next_item


class TestGetNextItem(TestCase):

    def get_lines(self, t):
        lines = t.split('\n')
        return [l + '\n' for l in lines]

    def test_basic(self):
        lines = self.get_lines(""" end_day (max_dom)                   = 12,	; two digit day of ending time
 end_hour (max_dom)                  = 12,	; two digit hour of ending time
 end_minute (max_dom)                = 00,	; two digit minute of ending time""")
        self.assertEqual(get_next_item(lines), 'end_day (max_dom)                   = 12,	; two digit day of ending time')

    def test_basic_2(self):
        lines = self.get_lines(""" end_second (max_dom)                = 00,	; two digit second of ending time
                                                  It also controls when the nest domain integrations end
                                                  All start and end times are used by real.exe.

                                                  Note that one may use either run_days/run_hours etc. or 
                                                  end_year/month/day/hour etc. to control the length of 
                                                  model integration. But run_days/run_hours
                                                  takes precedence over the end times. 
                                                  Program real.exe uses start and end times only.

 interval_seconds                    = 10800,	; time interval between incoming real data, which will be the interval
                                                  between the lateral boundary condition file
 input_from_file (max_dom)           = T,       ; whether nested run will have input files for domains other than 1
 fine_input_stream (max_dom)         = 0,       ; field selection from nest input for its initialization
                                                  0: all fields are used; 2: only static and time-varying, masked land 
                                                  surface fields are used. In V3.2, this requires the use of 
                                                  io_form_auxinput2""")
        self.assertEqual(get_next_item(lines), """end_second (max_dom)                = 00,	; two digit second of ending time
                                                  It also controls when the nest domain integrations end
                                                  All start and end times are used by real.exe.""")

    def test_basic_3(self):
        lines = self.get_lines(""" cycling                             = F,       ; whether this run is a cycling run, if so, initializes look-up table for Thompson schemes only
 restart_interval		     = 1440,	; restart output file interval in minutes
 reset_simulation_start              = F,       ; whether to overwrite simulation_start_date with forecast start time
 io_form_history                     = 2,       ; 2 = netCDF """)
        self.assertEqual(get_next_item(lines), 'cycling                             = F,       ; whether this run is a cycling run, if so, initializes look-up table for Thompson schemes only')

    def test_advanced(self):
        lines = self.get_lines("""For additional regional climate surface fields

 output_diagnostics                  = 1        ; adds 36 surface diagnostic arrays (max/min/mean/std)
 auxhist3_outname                    = 'wrfxtrm_d<domain>_<date>' ; file name for added diagnostics
 io_form_auxhist3                    = 2        ; netcdf
 auxhist3_interval                   = 1440     ; minutes between outputs (1440 gives daily max/min)
 frames_per_auxhist3                 = 1        ; output times per file
                                                  Note: do restart only at multiple of auxhist3_intervals

For observation nudging:
 auxinput11_interval                 = 10       ; interval in minutes for observation data. It should be 
                                                  set as or more frequently as obs_ionf (with unit of 
                                                  coarse domain time step).
 auxinput11_end_h                    = 6        ; end of observation time in hours""")
        self.assertEqual(get_next_item(lines), """output_diagnostics                  = 1        ; adds 36 surface diagnostic arrays (max/min/mean/std)""")

class TestConfigItem(TestCase):
    def test_basic_parse(self):
        i = ConfigItem(' time_step_fract_num                 = 0,	; numerator for fractional time step\n')
        self.assertFalse(i.is_section)
        self.assertEqual(i.name, 'time_step_fract_num')
        self.assertEqual(i.default, '0,')
        self.assertEqual(i.description, 'numerator for fractional time step')

    def test_multiline(self):
        i = ConfigItem(" e_vert (max_dom)                    = 30,	; end index in z (vertical) direction (staggered dimension)\n                                                  Note: this refers to full levels including surface and top\n                                                  vertical dimensions need to be the same for all nests\n                                                  Note: most variables are unstaggered (= staggered dim - 1)\n")
        self.assertFalse(i.is_section)
        self.assertEqual(i.name, 'e_vert')
        self.assertTrue(i.is_multi_dim)
        self.assertEqual(i.default, '30,')
        self.assertEqual(i.description, 'end index in z (vertical) direction (staggered dimension)\nNote: this refers to full levels including surface and top\nvertical dimensions need to be the same for all nests\nNote: most variables are unstaggered (= staggered dim - 1)')


from datetime import timedelta
from os.path import join
from collections import OrderedDict

import yaml

from wrfconf.serialize import dump
from wrfconf.utils import make_list, convert_str_to_dt, convert_dt_to_str, merge_dicts

def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
    class OrderedLoader(Loader):
        pass
    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))
    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)
    return yaml.load(stream, OrderedLoader)


def create_wrf_namelist(conf, wrf_dir):
    wrf_config = OrderedDict()
    run_info = conf['run_info']
    domain = conf['domain']
    start_time = convert_str_to_dt(run_info['start_date'])
    end_time = convert_str_to_dt(run_info['start_date']) + timedelta(hours=run_info['run_hours'])

    # Merge in the time info
    max_dom = run_info['max_dom']
    wrf_config['time_control'] = {
        'run_days': 0,
        'run_hours': run_info['run_hours'],
        'start_year': make_list(start_time.year, max_dom),
        'start_month': make_list(start_time.month, max_dom),
        'start_day': make_list(start_time.day, max_dom),
        'start_hour': make_list(start_time.hour, max_dom),
        'start_minute': make_list(start_time.minute, max_dom),
        'start_second': make_list(start_time.second, max_dom),
        'end_year': make_list(end_time.year, max_dom),
        'end_month': make_list(end_time.month, max_dom),
        'end_day': make_list(end_time.day, max_dom),
        'end_hour': make_list(end_time.hour, max_dom),
        'end_minute': make_list(end_time.minute, max_dom),
        'end_second': make_list(end_time.second, max_dom),
    }

    # Merge in the domain info
    domain_keys_to_copy = ('e_we', 'e_sn', 'dx', 'dy', 'parent_id', 'i_parent_start', 'j_parent_start', 'parent_grid_ratio')

    wrf_config['domains'] = {
        'max_dom': max_dom,
        'grid_id': list(range(1, max_dom + 1)),
        'parent_time_step_ratio': domain['parent_grid_ratio']
    }
    for k in domain_keys_to_copy:
        wrf_config['domains'][k] = domain[k]

    res = dump(merge_dicts(wrf_config, conf['namelist']))

    with open(join(wrf_dir, 'namelist.input'), 'w') as fh:
        fh.writelines(res)


def create_wps_namelist(conf, wps_dir):
    wrf_config = OrderedDict()
    run_info = conf['run_info']
    domain = conf['domain']
    end_time = convert_str_to_dt(run_info['start_date']) + timedelta(hours=run_info['run_hours'])

    # Merge in the time info
    max_dom = run_info['max_dom']
    wrf_config['share'] = {
        'max_dom': max_dom,
        'start_date': make_list(run_info['start_date'], max_dom),
        'end_date': make_list(convert_dt_to_str(end_time), max_dom),
    }

    # Merge in the domain info
    domain_keys_to_copy = ('e_we', 'dx', 'dy', 'e_sn', 'parent_id', 'parent_grid_ratio', 'i_parent_start', 'j_parent_start', 'ref_lat', 'ref_lon',
                           'ref_x', 'ref_y', 'truelat1', 'truelat2', 'stand_lon', 'geog_data_res')
    wrf_config['geogrid'] = {}
    for k in domain_keys_to_copy:
        if k in domain:
            wrf_config['geogrid'][k] = domain[k]

    res = dump(merge_dicts(wrf_config, conf['wps']))
    with open(join(wps_dir, 'namelist.wps'), 'w') as fh:
        fh.writelines(res)


def process_conf_file(conf_file, wrf_dir, wps_dir):
    try:
        with open(conf_file) as fh:
            conf = ordered_load(fh, yaml.SafeLoader)
    except IOError as e:
        raise Exception('Could not open file: {}'.format(conf_file))

    create_wrf_namelist(conf, wrf_dir)
    create_wps_namelist(conf, wps_dir)
